/* mPCIe-IDIO-8 VxWorks 7 (0600) "driver" Sample Program */

/* Copyright (c) 2019, ACCES I/O Products, Inc.
 * 10623 Roselle St, San Diego CA, 92121-1506 USA
 * +1-858-550-9559 | +1-800-326-1649 | sales@accesio.com
 * 
 * The right to copy, distribute, modify, or otherwise make use of this software 
 * may be pursuant to the terms of an applicable ACCES I/O Products, Inc. 
 * (ACCES) license agreement, or in the absence of such agreement, is limited to 
 * use with ACCES hardware products.
 */

/*
modification history
--------------------
1,apr,j2h    Forked for mPCIe-IDIO-8 from PCIe-DIO-24HS
13feb19,j2h  Initial release.
*/

/*
DESCRIPTION

Sample program for ACCES I/O Products IDIO cards driver.

*/

#include <vxWorks.h>
#include <private/iosLibP.h>
#include <hwif/buslib/pciDefines.h>
#include <stdio.h>
#include <string.h>
#include <usrLib.h>                     /* for devs() */
#include <hwif/vxBus.h>
#include <hwif/vxbus/vxbBusType.h>
#include <subsys/int/vxbIntLib.h>
//#include <hwif/buslib/vxbPciLib.h>

#include "vxbAIO_IDIO.h"

IMPORT void sysMsDelay (int);   /* delays int msecs */

#define NO_ARG_CALL_TEST(FNAME)                                     \
do { if (FNAME () == OK)                                            \
        printf (#FNAME "() OK.\n");                                 \
    else                                                            \
        {                                                           \
        printf (#FNAME "() failed with errno = %d.\n", errno);      \
        goto L_err;                                                 \
        }                                                           \
    }                                                               \
while (0)

int accesDioTestISRcount;

/******************************************************************************
*
* accesDioTestISR - simple ISR for interrupt test--just increment a count
*/

STATUS accesDioTestISR (void)
    {
    accesDioTestISRcount++;
    return OK;
    }

/******************************************************************************
*
* accesDioIntTest - test DIO card Change-of-State interrupts
* 
* Do all steps to enable interrupts, wait 30 seconds for the user to short a
* pin, and report # interrupts received. A quick paper-clip short will generate
* a few dozen interrupts. See code for details.
*/

STATUS accesDioIntTest (void)
    {
    /***** TEST: Interrupts. */ 

	accesDioTestISRcount = 0; 
	
    /* Establish our ISR callback function and enable interrupts for 30 seconds. */
    NO_ARG_CALL_TEST(accesDioOpen);
    
    if (accesDioIntCallbackInstall (accesDioTestISR) == ERROR)
        { printf ("accesDioIntTest:%d: failed with errno = %d.\n", __LINE__, errno); goto L_err; }

    /* Remember! Ports must be set for input for interrupts to work. */
    if (accesDioConfigure(PORT_DIR_INPUT, PORT_DIR_INPUT) == OK) 
        printf ("accesDioConfigure() OK\n");
    else
        { printf ("accesDioTest:%d: failed with errno = %d.\n", __LINE__, errno); goto L_err; }

    accesDioIntClear();
    
    if (accesDioIntEnable () == ERROR)
        { printf ("accesDioIntTest:%d: failed with errno = %d.\n", __LINE__, errno); goto L_err; }

    accesDioIntClear();
    
    printf ("Connect a digital transition source to any digital input and cause a known\n"
    		"number of edges to be generated\n"
            "Our callback function, accesDioTestISR(), increments a local counter which\n"
            "is reported at the end of the test.\n"
    		"T13 Verify reported count matches generated number of edges.\n");
    
      
    sysMsDelay (10000);
    
    NO_ARG_CALL_TEST(accesDioClose);
    
    printf ("Interrupt test done. accesDioTestISRcount = %d\n", accesDioTestISRcount);

    return OK;
    
    /* In Workbench, place a breakpoint here to break on errors. */
L_err:
    printf ("ERROR: At L_err label. errno if relevant = %d.\n", errno);
    return ERROR;
    }

/******************************************************************************
*
* AIO_Sample - demonstrate IDIO driver functions
* 
*/

STATUS AIO_Sample (void)
    {
    UINT8  TTLOut, TTLIn, ISOIn, FETOut;
    int    cycle;
    UINT32 pinCnt;
    
    printf("This sample runs through some common operations to demonstrate the API\n\n");
    
    printf("Opening driver\n");
    if (accesDioOpen() != OK)
    	{ printf ("accesDioOpen() failed @ line %d with errno = %d.\n", __LINE__, errno); goto L_err; }    	
    
    printf("Configuring TTL 0-3 for output, 4-7 for input\n");
    if (accesDioConfigure( PORT_DIR_OUTPUT, PORT_DIR_INPUT) != OK)
        { printf ("accesDioConfigure failed @ line %d with errno = %d.\n", __LINE__, errno); goto L_err; }
    
    printf("Writing 0xAA to TTL (writes to inputs are ignored)\n");
	accesIdioWriteTTL(0xAA);
	
	printf("Reading TTL bits\n");
	accesDioReadPorts(NULL, &TTLIn);
    printf ("TTL Data Written = 0xAA\n"
    		"TTL Data Read = 0x%02hhx\n"
    		"Expected 0xAA with wrap plug, 0xFA without\n", 
			TTLIn);

    printf("Fast toggle on TTL bits 0-3: check w/o'scope\n");
    for (cycle = 1; cycle <= 1000000; cycle++)
    {
    	accesIdioWriteTTL(0xFF);
    	accesIdioWriteTTL(0x00);
    }
    
    printf("TTL LED fixture displays 1 walking bit on TTL bits 0-3, twice\n");
    for (cycle = 1; cycle <= 2; cycle++)
        {
        TTLOut = 1;
        for (pinCnt = 1; pinCnt <= 4; pinCnt++)
            {
            accesIdioWriteTTL (TTLOut);
            TTLOut = (UINT8)(TTLOut << 1);
            sysMsDelay (100);
            }
        
        for (pinCnt = 4; pinCnt >= 1; pinCnt--)
            {
            accesIdioWriteTTL (TTLOut);
            TTLOut = TTLOut >> 1;
            sysMsDelay (100);
            }
        }
    
    printf("Isolated Inputs\n");
    accesDioReadPorts(&ISOIn, NULL);
    printf("ISO bits read %02hhx\n", ISOIn);
    
    printf("FET walking bit\n");
    for (cycle = 1; cycle <= 2; cycle++)
        {
    	FETOut = 0x01;
        for (pinCnt = 1; pinCnt <= 8; pinCnt++)
            {
            accesIdioWriteFET (FETOut);
            FETOut = (UINT8)(FETOut << 1);
            sysMsDelay (100);
            accesDioReadPorts(&ISOIn, NULL);
            printf("ISO bits read %02hhx\r", ISOIn);
            }
        FETOut = 0x80;
        for (pinCnt = 8; pinCnt >= 1; pinCnt--)
            {
            accesIdioWriteFET (FETOut);
            FETOut = FETOut >> 1;
            sysMsDelay (100);
            accesDioReadPorts(&ISOIn, NULL);
            printf("ISO bits read %02hhx\r", ISOIn);
            }
        }    
    
    
    printf("\nClosing driver.\n");
    if (accesDioClose() != OK)
    	{ printf ("accesDioClose() failed @ line %d with errno = %d.\n", __LINE__, errno); goto L_err; }    	
  
    
    printf ("Done.\n");
    return OK;
    
    /* In Workbench, place a breakpoint here to break on errors. */
L_err:
    printf ("ERROR: At L_err label. errno if relevant = %d.\n", errno);
    accesDioClose();
    return ERROR;
    }


