/* vxbAIO_IDIO.c - ACCES I/O Products IDIO driver */

/* Copyright (c) 2018, 2019 ACCES I/O Products, Inc.
 * 10623 Roselle St, San Diego CA, 92121-1506 USA
 * +1-858-550-9559 | +1-800-326-1649 | sales@accesio.com
 *
 * The right to copy, distribute, modify, or otherwise make use of this software
 * may be pursuant to the terms of an applicable ACCES I/O Products, Inc.
 * (ACCES) license agreement, or in the absence of such agreement, is limited to
 * use with ACCES hardware products.
 */

/*
modification history
--------------------
12mar19,j2h  continued changes towards mPCIe-IDIO-8 version
25feb19,j2h  forked for mPCIe-IDIO-8
21dec18,j2h  Adding AIOWDM.dll equivalent API, changed kprintf to printf in accesDioShow, modified for PCIe-DIO-24HS
13jun18,acc  Release 1.0.
*/

/*
DESCRIPTION
===========
This driver provides installation (probe and attach) and I/O routines for
selected ACCES I/O Products Digital I/O cards, including interrupt support
(example: mPCIe-DIO-24S). These cards emulate aspects of the 8255, in some
cases with extensions such as interrupt support.

This "Description" augments that found in the companion header file
vxbAccesDio.h. It is recommended that that file be read before continuing here.
Note that it includes "References" for the DIO card, the 8255, and PCI.

*** NOTE: This version of this driver supports only a single ACCES DIO card in
*** a system.

This driver is for use on a PCI bus. Supported registers are at BAR[2] offsets
in hex of:
  +00: FET 0-7 Output / readback
  +01: Isolated Inputs

  +10: TTL0-3 & TTL 4-7
  +11: TTL I/O Group Direction Control

  +28: IRQEN:   Write 0xFF to enable interrupts, 0x00 to disable interrupts.
  +29: IRQCLR : Write 0xFF to clear all pending IRQs.
  +29: IRQSTAT: On read, if bit 0 is set, an interrupt is pending.

NOTE: BAR[2] is I/O type, not memory-mapped. See details in the ACCES hardware
manual cited in the References in the header file.

Include Files
-------------
vxbAIO_IDIO.h (include to access DIO-specific open, close, read, etc.).

User-Callable Functions
-----------------------
For details of user-callable functions, see comments with each. All return
STATUS and if returning ERROR, set global errno. See each for details.

accesDioClose( )            close the card, releasing it for another user
accesDioConfigure( )        set the input vs. output for each port
accesDioIntCallbackInstall( ) set the user ISR callback function pointer
accesDioIntClear( )         clear interrupts; normally done by the ISR
accesDioIntDisable( )       disable interrupts
accesDioIntEnable( )        enable interrupts; see it for other requirements
accesDioIntPending( )       return TRUE if was generated but not cleared
accesDioOpen( )             open the card for exclusive access
accesDioReadPorts( )        read from all ports
accesDioShow( )             show the state of the driver and device
accesDioWritePorts( )       write to all ports
accesDioGetDriverRevision() retrieve the driver revision and build date/time

Additional Functions for Internal Use
-------------------------------------
The following functions are not intended for direct use by an ordinary user.
Calling them in a test program or from the VxWorks kernel shell may be of use
to a driver developer. If ACCES_DEBUG is defined, LOCAL will resolve to empty
and the functions will be visible globally. They are not included in any
header file and must be IMPORTed explicitly.

accesDioCloseForce( )       close the card with no error checks
accesDioDebugLevelSet( )    set debug output verbosity level; 0 means none
accesDioInt( )              service interrupts (the ISR)
accesDioPciAttach( )        attach function for PCI bus
accesDioPciDrvInit( )       initialize the driver
accesDioPciProbe( )         probe function for PCI bus
accesDioRead32Raw( )        read the 8255 registers with no masking, no checks
accesDioReadControl( )      read the Control Word register, no error checks
accesDioWrite8Raw( )        write a byte to a register, checks only reg offset
accesDioWriteControl( )     write the Control Word register, no error checks
accesDioWritePortsRaw( )    write the ports (not the Control Word); no checks
isEntryError( )             check entry conditions for user-callable functions
isDriverOK( )               check that the driver appears to be installed

Implementation Summary
======================
NOTE: As an aid to development of future ACCES I/O Products drivers for VxWorks,
this driver includes extensive implementation notes.

**** CAUTION: VxWorks is very flexible and there are often several ways to code
**** a feature. Consult ACCES I/O Products or VxWorks 7 documentation for
**** alternatives. The approach taken here is based on VxWorks 7 release SR0530
**** 2018-03-16.

For an introduction to the VxWorks 7 I/O subsystem, drivers, and devices, see
the "VxWorks(R) 7 Programmer's Guide," Edition 19, 29 March 2018, chapters 11
"I/O System" and 12 "Devices", especially chapter 12 section "I/O Driver Model
(335-355). The implementation notes in this file augment that section for this
specific driver; reading that section first is recommended.

Supporting a typical VxWorks VxBus-based device requires four high-level steps
(further details under "Driver Definition and Installation" below):

1. Only once, a driver for the type of device is declared statically with
   required data with a VXB_DRV_DEF macro. Top-level PCI bus setup code uses
   this data to call vxbDrvAdd() to register the driver with VxBus. That data
   typically includes pointers to the PCI "probe" and "attach" functions used in
   steps 2 and 3 next.

2. During installation of a PCI bus, the top-level PCI bus setup code scans the
   bus and for each and every device found, calls the driver's "probe" function
   for every PCI driver registered in step 1. If a driver supports that device,
   it returns OK. (See vxbLib.c:vxbDevProbe() for "best match" refinements.)

3. If the probe is successful, the PCI bus setup code then calls that driver's
   "attach" function. That function calls (directly or indirectly), vxbDevAdd()
   to add the device to the I/O subsystem, and performs other device-specific
   actions such as saving BAR* information, connecting and enabling interrupts,
   setting the baud rate for serial devices, etc.

       *BAR: Base Address Register. The virtual address of the device in "PCI
       Configuration Space"; see the Wikipedia source and its references given
       in the header file.

4. Only once, the driver itself calls iosDrvInstall() to provide callback
   functions for user-callable, high-level open(), close(), read(), write(),
   ioctl(), create(), and remove() functions. These high-level functions are
   NOT supported by this driver, which has its own user-callable DIO-specific
   functions. Nevertheless, iosDrvInstall() is called in this driver for two
   reasons:

   - The call to iosDrvInstall() passes NULL pointers for all of the callback
     functions so that user attempts to make the high-level calls return ERROR.

   - The devs command to the VxWorks kernel shell displays all devices and the
     driver number for each. We want ACCES cards to appear in that display and
     the call to iosDrvInstall() returns that driver number (it is the index to
     the driver callbacks table).

   The code to make this call to iosDrvInstall() may be part of:

   a) An initialization function in the driver called during VxWorks startup
      due to an INIT_RTN statement in the Component Description File (CDF) for
      the driver. However, this approach requires understanding the order in
      which initialization is carried out (because, for example, iosInit() must
      be called by VxWorks initialization code before any other IOS functions
      can be called, including iosDrvInstall()).
   b) The driver's Probe function after testing that that it has not already
      made the call. This is sub-optimal because the Probe function is called
      many more times than the Attach function.
   c) The driver's Attach function after testing that that it has not already
      made the call. This simple approach is taken here.

Two critical data structures support the above:

1. typedef struct vxbDev {...} VXB_DEV
      and the associated
   typedef struct vxbDev * VXB_DEV_ID

   The vxbDev structure is allocated in step 2 above by the top-level PCI bus
   setup code and used to call the driver "probe" and "attach" functions. It is
   defined in
   $WIND_BASE/pkgs/os/drv/vxbus-<version>/core<-version>/src/vxbDevDef.h and
   includes a member for the device's name (pName), a pointer through which
   the "probe" can send data to the "attach" (pVxbSoftc), a pointer to the driver
   data of step 1 above (pVxbDriver), and a pointer to the driver table entry
   where the callback function pointers of step 4 above are stored.

2. The "device descriptor" structure. This all-important structure is described
   in detail below in section "DIO Device Descriptor and Pointer to It". For most
   devices supported by VxWorks, a device descriptor is dynamically allocated
   for each instance of the device. In this version of this driver, a single DIO
   card is supported and so a module LOCAL static structure is used.

Key implementation steps:
- Top-level PCI bus enumeration code calls our accesDioPciProbe(), which returns
  OK when the device passed to it "matches" a device the driver supports.
- As a result, for that device, the top-level PCI bus enumeration code calls our
  accesDioPciAttach() function. It calls iosDrvInstall() as described above,
  sets a module LOCAL pointer to our single LOCAL device descriptor structure
  accesDioDd, gets and saves in the device descriptor the "handle" and BAR
  information necessary to access the device, sets up to receive interrupts for
  the device and finally, calls iosDevAdd() to add our DIO device to the I/O
  subsystem.

At that point, a user task can call our specific user-callable functions listed
above to open/close, read/write, and control interrupts for the device.

Additional major notes sections below are:
- "Debugging": describes debugging features included by defining the ACCES_DEBUG
  macro.
- "Local defines", including register BAR offsets for the individual ports.
- "DIO Device Descriptor and Pointer to It".
- "Driver Definition and Installation": a general description of adding a driver
  to VxWorks 7, using this driver as an example.
*/

/* includes */

#include <vxWorks.h>
#include <hwif/vxBus.h>
#include <hwif/buslib/vxbPciLib.h>
#include <errno.h>
#include <ioLib.h>
#include <iosLib.h>
#include <private/kwriteLibP.h>
#include <stdio.h>
#include <sysLib.h>
#include <taskLib.h>
#include "vxbAIO_IDIO.h"


/*******************************************************************************
Debugging
---------
This section define macros and module-level variables to include and control
debug code and messages. If ACCES_DEBUG is defined:

- Define macro DBG_MSG (level, fmt, ...), which uses kprintf () to display a
  message if accesDioDebugLevel > level.

- Define convenience macros DBG_ENTER, DBG_EXIT_OK, DBG_MSG, and DBG_MSG0--see
  below.

- Define LOCAL (defined by VxWorks as static) as nothing so that all normally
  private variables and functions in this file are global and can be examined in
  a debugger or using the VxWorks kernel shell.

- In accesDioPciProbe(), sets _func_kwrite() (used by kprintf()) to kwriteVga().
  This enables use of DBG_MSG () in all functions in this driver without
  exception (kwriteVga() outputs to the PC Console with ~no initialization).

- Define and initialize accesDioDebugLevel; valid values are:
  0: No debug output, even if compiled in because ACCES_DEBUG is defined.
  1: Minimal: message at each function entries and exit (with return value).
  2: Verbose: 1 + additional information during function execution.
  3: Bypass the Probe (and therefore the Attach) function during VxWorks
     initialization; intent is to call each manually or in a test program
     executed from the VxWorks kernel shell.

  *****************************************************************************
  * CAUTION: If interrupts are enabled, generating debug output may cause the *
  * system to crash after a few interrupts. During testing, to avoid this,    *
  * set accesDioDebugLevel to 0 or call accesDioDebugLevelSet(0) after boot   *
  * up (e.g., from the VxWorks kernel shell).                                 *
  *****************************************************************************

- Define and initialize module LOCAL accesDioProbeFail. If set to TRUE and
  accesDioDebugLevel >=3, our PCI bus Probe function (accesDioPciProbe()) will
  always return ERROR so that the DIO device is NOT discovered during normal
  VxBus PCI enumeration. Can then call Probe and Attach after VxWorks is up.

- Define accesDio_pVxbDev. This is a local copy of the argument passed to our
  accesDioPciProbe() and accesDioPciAttach()) functions to enable it to be
  examined in the Workbench debugger or VxWorks kernel shell.
*/

#undef  ACCES_DEBUG
 
// #define ACCES_DEBUG      Comment out to remove debug code. 


#ifdef ACCES_DEBUG

STATUS kwriteVga (char * pBuffer, size_t len);
int kprintf(const char * fmt, ... );

#ifndef INCLUDE_EARLY_PRINT

#endif
/* Make local (private) symbols global. */

#ifdef  LOCAL
#undef  LOCAL
#define LOCAL
#endif /* ACCES_DEBUG */

/* DBG_MSG messages--includes function name and line. */
#define DBG_MSG(level, fmt, ...)                                                \
    do                                                                          \
        if ( accesDioDebugLevel >= level )                                      \
                kprintf("%s:%d: " fmt, __func__, __LINE__, ##__VA_ARGS__);      \
    while (0)

/* DBG_MSG0 messages--no function name and line. */
#define DBG_MSG0(level, fmt, ...)                                               \
    do                                                                          \
        if ( accesDioDebugLevel >= level )                                      \
			kprintf(fmt, ##__VA_ARGS__);                                        \
    while (0)

/* Use DBG_ENTER as the first statement of every function. */
#define DBG_ENTER     DBG_MSG(1, "Enter.\n")

/* Use DBG_EXIT_OK as the last statement of a function before "return OK;". */
#define DBG_EXIT_OK   DBG_MSG(1, "Exit OK.\n")

/* Variables to control debug; see details above. */
LOCAL int           accesDioDebugLevel = 9;     /* See valid values above. */
LOCAL BOOL          accesDioProbeFail = FALSE;  /* See explanation above. */
LOCAL VXB_DEV_ID    accesDio_pVxbDev;           /* So can examine in shell. */

/*
 * In accesDioPciDrvInit(), will set up to use kwriteVga() to support kprintf().
 * Must explicitly declare kwriteVga() because it is not included in any header
 * file.
 */

#else   /* !ACCES_DEBUG */
#define DBG_MSG(level, fmt, ...)
#define DBG_MSG0(level, fmt, ...)
#define DBG_ENTER
#define DBG_EXIT_OK
#endif  /* ACCES_DEBUG */

/*******************************************************************************
* Local macros.
*/

/* Use EXIT_ERR as the last statement of a function before "return ERROR;"
 * for a simple error exit that only sets errno and reports it. */
#ifndef ACCES_DEBUG
#define EXIT_ERR(errnoVal)                      \
    errno = errnoVal
#else
#define EXIT_ERR(errnoVal)                      \
    do                                          \
        {                                       \
        errno = errnoVal;                       \
        DBG_MSG (1, "Exit ERROR %d\n.", errno); \
        }                                       \
    while (0)

#endif /* ACCES_DEBUG */

/*******************************************************************************
* Local declarations.
*/

/*
 * Register identifiers = BAR offsets; use in calls to accesDioWrite8Raw().
 * Define as an enum for readability in Workbench, but specify a value for each
 * enumeration constant equal to its BAR offset and rely on that in code.
 *
 * See accesDioIntEnable() for details for REG_COS_MASK, REG_IRQ_EN, and
 * REG_IRQ_CLR.
 */

typedef enum reg
    {
    REG_FET         = 0x00,
    REG_ISODIG      = 0x01,
    REG_TTLDATA     = 0x10, /* each nybble is an I/O Group (0-3 & 4-7) */
    REG_TTLCONTROL  = 0X11,
    REG_IRQ_EN      = 0x28, /* Write 0xff to enable ints, 0x00 to disable. */
    REG_IRQ_CLR_STAT= 0x29  /* Write 0xff t0 clear pending ints; read for */
    } REG;                  /* status: 0x1 = IRQ generated and not cleared. */

/*
DIO Device Descriptor and Pointer to It
---------------------------------------
struct accesDioDd, type ACCES_DIO_DD, pointer pAccesDioDd

accesDioDd is a data structure that contains all required information to use an
ACCES DIO card once it has been added to the I/O system.

The first member is a DEV_HDR structure as is required for all VxBus devices.
The first member of a DEV_HDR is a list pointer that links the device into the
global device list (used, for example, by the shell devs command to display a
list of devices).

For most devices, the device descriptor is dynamically allocated by a device
"create" or "add" function--called usually in the case of PCI devices by the
device's Attach function (or a function it calls), which in turn calls
iosDevAdd() with three arguments: a pointer to the device descriptor, the
device's name, and the driver number for the device (see accesDioPciAttach
here).

Because this version of this DIO driver supports only a single card in a system
(presented as one device--as opposed to a device per pin or port), the
accesDioDd struct is a module LOCAL variable in this file. However to prepare
to support multiple DIO devices in a system, references to its members are
through module LOCAL pointer pAccesDioDd rather than directly using using dot
notation.

NOTE: a device descriptor was historically called a "device control structure";
the term "device descriptor" is from the "VxWorks 7 Programmer's Guide," chapter
12 "Devices," "I/O Driver Model," "Devices" (p. 341).

Notes for individual members in the device descriptor:

- taskId:
    To have exclusive access after calling accesDioOpen() until accesDioClose(),
    need way to "lock" the card for the current user (meaning return ERROR on
    calls from other tasks). Use the taskId to store the task ID of the calling
    task on accesDioOpen(). Only requests from that task will be permitted.

- handle and bar:
    The handle is used to access the actual device registers, whether memory-
    mapped or in I/O register address space type (one member of the handle is
    the type--see pAdfRes below). The device registers are relative to a
    (virtual) base address, and that address is stored in *part* of a 32-bit
    "Base Address Register" (BAR). A device may use up to 6 BARs and a device
    driver and the device maker must agree on the use of the BARs.

        For the ACCES DIO cards, BAR[2] is I/O type and is the base for the
        port, Control Word, and interrupt-related registers. BAR[0] (also
        I/O type) and BAR[1] (memory type) are for other purposes not supported
        by this driver; see the appropriate ACCES Hardware Manual for details.

    Thus accessing a register requires specifying three values: a handle,
    a BAR, and an offset from the BAR. The offset is hard-coded as the reg
    enumeration above based on information in the hardware manual.

    (VxWorks provides six functions to read and write a device register
    of 8, 16, or 32 bits: vxbRead8(), vxbWrite8(), ... vxbWrite32() (declared
    in vxbAccess.h; 64-bit versions are provided for some architectures). The
    first two parameters for all vxbReadXX/vxbWriteXX functions are the BAR
    and the offset.

- pAdrRes:
    A pointer to an "address resource" used in getting the "handle" to the
    card (the handle introduced just above.)

    A VxWorks Bus "resource" provides access to devices. There are three types
    of resource: 1) a memory-mapped device address, 2) a device address in I/O
    address space for systems with port "in" / "out" instructions (x86), and
    3) an interrupt. A resource virtualizes each of these so that a reads and
    writes to a device need not be concerned about the hardware details.

    To obtain a resource, the device driver allocates a resource of a particular
    type. For our DIO two resource allocations are required (both done in our
    Attach function):

            pAdrRes = vxbResourceAlloc (pVxbDev, VXB_RES_IO, 2);
            pIntRes = vxbResourceAlloc (pVxbDev, VXB_RES_IRQ, 0);

    vxbResourceAlloc() allocates a small memory block and sets it based on the
    resource type.

    The first argument for both points to the device node in the complete list
    of devices added to the system. It is either allocated by the user or, in
    for PCI devices, allocated by the VxBus code that enumerates the devices on
    the bus, and is then passed to the device driver probe and attach functions
    to identify the discovered device.

    The second argument is the resource type: VXB_RES_IO, VXB_RES_MEM, or
    VXB_RES_IRQ to specify the type of resource to allocate.

    The third argument identifies the resource to allocate depending on the
    type.
    - VXB_RES_IO and VXB_RES_MEM
        The desired Base Address Register (BAR); a PCI bus device has up to six
        BAR of either type, numbered 0-5. Thus, the goal of the first call above
        is to get a handle to BAR[2], which is the BAR from which the
        mPCIe-DIO-24S registers are offset.
    - VXB_RES_IRQ: device-dependent, usually 0.

    Once the resource is allocated, for memory and I/O resources, the handle
    may be indirectly obtained from the resource--see the code.

    Technically, once the handle has been obtained, this address resource
    could be immediately freed. It is instead saved in our device descriptor
    for two reasons:
    - In device descriptors allocated dynamically, if the device is later
      deleted, the resource allocation must also be deleted.
    - To aid in debugging.

-pIntRes:
    pIntRes points to the resource allocated to connect to the interrupts from
    the card by calling vxbIntConnect() (in our Attach function). See pAdrRes
    for general details. Unlike the address resource which could be freed after
    the handle is obtained, the interrupt resource must be for as long as
    interrupts from the device are to be enabled. Thus, it is saved in our
    device descriptor for that reason as well as the two reasons given for
    pAdrRes.
*/

typedef struct accesDioDd
    {
    DEV_HDR         devHdr;             /* I/O device header; includes device
                                           name and driver number */
    char *          pProductName;       /* ACCES product name; set by
                                           accesDioAttach() to the name in the
                                           DIO_PCI_DEV_CFG structure
                                           (see "Match Table" below); example:
                                           "mPCIe-DIO-24S" */
    TASK_ID         taskId;             /* ID of task card was opened by; set by
                                           accesDioOpen(); NULL means closed.*/
    void *          handle;             /* see "handle" above */
    ULONG           bar;                /* BAR virtual address; see above */
    STATUS          (*pIntCallback)();  /* for pointer to user function to call
                                           on an interrupt */
    BOOL            intEnabled;         /* TRUE if interrupts are enabled */
    VXB_RESOURCE *  pAdrRes;            /* address resource described above */
    VXB_RESOURCE *  pBAR0Res;           /* added 21dec18 in debug attempt */
    void *          handle0;
    ULONG           BAR0;
    VXB_RESOURCE *  pIntRes;            /* interrupt resource described above */
    int             intCount;           /* see accesDioInt() */

    } ACCES_IDIO_DD;


/* Device descriptor and pointer:
 * Declare our module-level static device descriptor and the pointer to it. The
 * pointer is set in accesDioPciAttach().
 */

LOCAL ACCES_IDIO_DD    accesDioDd;
LOCAL ACCES_IDIO_DD *  pAccesDioDd;


/*******************************************************************************
Driver Definition and Installation
==================================
Defining a driver and adding it to the I/O system requires several steps,
including meeting consistency requirements described below. The steps are
presented in top-down order here, but are implemented bottom-up to reduce
the need for forward references.

1. Declarations: Three declarations are required (illustrated using this ACCES
   DIO driver):

1.a. VXB_DRV_DEF (AccesDioPciDrv)

    Add the driver to the system. Calls (directly or indirectly depending on
    the type of image):

        vxbDrvAdd (VXB_DRV * pDriver [=&AccesDioPciDrv in 1.b]);

    vxbDrvAdd() registers the driver with the VxBus subsystem by linking the
    VxBus driver structure passed to it into the list of drivers rooted at
    global variable vxbDrvList (in vxbLib.c). vxbDrvAdd() then calls the Probe
    and Attach functions for any devices detected on the bus and having a
    matching class (as given by pDriver->vxbClass (VXB_BUSID_PCI here).

1.b. LOCAL VXB_DRV AccesDioPciDrv =
        {
        { NULL },               /@ SL_NODE          vxbNode: for drivers list @/
        "ACCES DIO PCI Driver", /@ char *           driver name @/
        "Driver for ACCES ...", /@ char *           description @/
        VXB_BUSID_PCI,          /@ vxb_BUSTYPE_ID   vxbClass        @/
        0,                      /@ UINT32           vxbFlags        @/
        0,                      /@ int              RefCnt          @/
        accesDioPciDrvMethods,  /@ VXB_DRV_METHOD * pVxbMethods--ptr to the
                                                    methods table @/
        NULL                    /@ void *           pVxbParamDefaults @/
        };

    Declare and initialize a ` static VXB_DRV structure as required by the
    VXB_DRV_DEF statement above. As noted, the structure will be linked into the
    global vxbDrvList. The structure contains a pointer to an array of callback
    functions specific to the driver--including generally for PCI devices the
    Probe and Attach functions--here accesDioPciProbe() and accesDioPciAttach()
    in this file.

1.c. VXB_DRV_METHOD accesDioPciDrvMethods[] = { ... }

    Declare and initialize the "Methods Table" array of callback functions
    required by the VXB_DRV accesDioPciDrv structure described above.

2. Executable statements

Registering and installing a driver with the VxBus subsystem requires two
function calls:

- vxbDrvAdd (VXB_DRV * pDriver)
  As noted above, this is called indirectly as a result of the VXB_DRV_DEF()
  macro above.

- iosDrvInstall() is called to set callbacks for seven basic I/O functions; as
  noted above, iosDrvInstall() is called in this driver but with all NULL
  pointers for the callback functions (see "Implementation Summary" step 4).

  At that point, devices for the driver may be added to the I/O subsystem,
  typically by calling a "create" or "add" function. For PCI devices, this is
  usually done in the Attach function (or a function it calls) as it is here.

Consistency requirements
------------------------
- Methods Table: The Methods Table member of the VXB_DRV structure (pVxbMethods)
  must point to the Methods Table in this file, here accesDioPciDrvMethods. See
  the VXB_DRV AccesDioPciDrv structure below.
*/

/* Implementation of above, with reference paragraph numbers. */

/* 1.c Methods Table: Declare the table of VxBus PCI methods implemented in this
 * file. See above for purpose. See vxbMethod.h for macros VXB_DEVMETHOD_CALL
 * and VXB_DEVMETHOD_END. */

/* Forwards for our Attach and Probe functions required for the Methods table. */
LOCAL STATUS accesDioPciAttach  (VXB_DEV_ID pDev);
LOCAL STATUS accesDioPciProbe   (VXB_DEV_ID pDev);

LOCAL VXB_DRV_METHOD accesDioPciDrvMethods[] =
    {
        { VXB_DEVMETHOD_CALL (vxbDevProbe),  accesDioPciProbe  },
        { VXB_DEVMETHOD_CALL (vxbDevAttach), accesDioPciAttach },
        VXB_DEVMETHOD_END
    };

/* 1.b VxBus driver structure. See "Driver Definition and Installation" above for
 * details. */

VXB_DRV AccesDioPciDrv =
    {
    { NULL },               /* SL_NODE          vxbNode: for drivers list */

    "ACCES IDIO PCI Driver", /* char * */
    "Driver for ACCES I/O Products IDIO for mPCIe bus" ,
                            /* char *           description */
    VXB_BUSID_PCI,          /* VXB_BUSTYPE_ID   vxbClass */
    0,                      /* UINT32           vxbFlags (see below) */
    0,                      /* int              RefCnt   */
    accesDioPciDrvMethods,  /* VXB_DRV_METHOD * pVxbMethods; Methods Table ptr */
    };

    /* vxbFlags: May be used to pass parameters to a driver as described in
     * pkgs/os/drv/vxbus-<version>/core-<version>/src/vxbParamLib.c. It is
     * invoked by providing a vxbFlags of VXB_DRVFLAG_PARAM instead of 0. When
     * used, the VXB_DRV structure has a final member pointing to a table of
     * parameter defaults. See vxbParamLib.c for details. */

/* 1.a Define the driver! Indirectly calls vxbDrvAdd(). See "Driver Definition
 * and Installation" for details. */

VXB_DRV_DEF (AccesDioPciDrv);


/*
 * drvNum: Declare the module LOCAL driver number returned by iosDrvInstall()
 * called in accesDioPciAttach(). Special values:
 *  0:          Driver not yet initialized with the iosDrvInstall call.
 * -1 (ERROR):  Driver initialization, Probe, or Attach failed.
 */

LOCAL int drvNum = 0;   /* Initialize to 0 to be explicit. */


/*
* Match Table and Configuration Data
* ----------------------------------
* Match Table: The Match Table is an array of VXB_PCI_DEV_MATCH_ENTRY structs.
* There must be one unique entry in this table for each PCI device supported by
* this driver as identified by the Vendor ID and Device ID of the device. The
* table is searched by our Probe function each time it is called by the PCI
* VxBus subsystem while enumerating the devices on the a PCI bus (see "Driver
* Definition and Installation" above for more). A match means this driver
* supports the device.
*
* The first two elements of the table are the Device ID and Vendor ID of a
* device supported by this driver. The third element of each table entry is a
* pointer to configuration data for that device. That data is defined by a local
* DIO_PCI_DEV_CFG structure. The pointer to that structure for the device
* matched during a probe is communicated from our Probe to our Attach
* function through the pVxbDrvData pointer in the VXB_DEV structure passed to
* Attach.
*
* The configuration structure type and instances are given first followed by
* the Match Table.
*/

/* Configuration data structure type. At present, just the product name. */
typedef struct dioDrvDevCfg
    {
    char *  pProductName;       /* Product name of ACCES card. */
    /* --J2H TODO: add fields similar to AIOWDM Registry "scripts" for IRQ clear/enable/disable  */
    } DIO_PCI_DEV_CFG;


/* Configuration data instances, one per device type to be recognized. */
LOCAL const DIO_PCI_DEV_CFG mPcie_IDIO_8_Cfg =
    {
    "mPCIe-IDIO-8"
    };
LOCAL const DIO_PCI_DEV_CFG mPcie_IIRO_8_Cfg =
    {
    "mPCIe-IIRO-8"
    };
LOCAL const DIO_PCI_DEV_CFG mPcie_IDIO_4_Cfg =
    {
    "mPCIe-IDIO-4"
    };
LOCAL const DIO_PCI_DEV_CFG mPcie_IIRO_4_Cfg =
    {
    "mPCIe-IIRO-4"
    };
LOCAL const DIO_PCI_DEV_CFG mPcie_IDO_8_Cfg =
    {
    "mPCIe-IDO-8" 
    };
LOCAL const DIO_PCI_DEV_CFG mPcie_RO_8_Cfg =
    {
    "mPCIe-RO-8"  
    };
LOCAL const DIO_PCI_DEV_CFG mPcie_II_16_Cfg =
    {
    "mPCIe-II-16" 
    };
LOCAL const DIO_PCI_DEV_CFG mPcie_II_8_Cfg =
    {
    "mPCIe-II-8"  
    };
LOCAL const DIO_PCI_DEV_CFG mPcie_II_4_Cfg =
    {
    "mPCIe-II-4"  
    };
LOCAL const DIO_PCI_DEV_CFG mPcie_QUAD_4_Cfg =
    {
    "mPCIe-QUAD-4"
    };
LOCAL const DIO_PCI_DEV_CFG mPcie_QUAD_8_Cfg =
    {
    "mPCIe-QUAD-8"
    };
    
/* Match Table */
LOCAL const VXB_PCI_DEV_MATCH_ENTRY matchTable[] =
    {
    /*   Device   Vendor  DIO_PCI_DEV_CFG instance
           ID       ID    pointer
          ------  ------  -------------------------- */
        { 0x0101, 0x494f, (void *)&mPcie_IDIO_8_Cfg },
        { 0x0102, 0x494f, (void *)&mPcie_IIRO_8_Cfg },
        { 0x0103, 0x494f, (void *)&mPcie_IDIO_4_Cfg },
        { 0x0104, 0x494f, (void *)&mPcie_IIRO_4_Cfg },
        { 0x0105, 0x494f, (void *)&mPcie_IDO_8_Cfg },
        { 0x0106, 0x494f, (void *)&mPcie_RO_8_Cfg },
        { 0x0107, 0x494f, (void *)&mPcie_II_16_Cfg },
        { 0x0108, 0x494f, (void *)&mPcie_II_8_Cfg },
        { 0x0109, 0x494f, (void *)&mPcie_II_4_Cfg },
        { 0x010a, 0x494f, (void *)&mPcie_QUAD_4_Cfg },
        { 0x010b, 0x494f, (void *)&mPcie_QUAD_8_Cfg },

        {} /* Empty terminates list */
    };


/* Forwards (continued) */
void        accesDioCloseForce (void);
STATUS      accesDioIntCallbackInstall (STATUS (*callback_funcptr)());
void        accesIdioWriteTTLRaw (UINT8 ports);
LOCAL BOOL  isDriverOK (void);
LOCAL BOOL  isEntryError (ACCES_IDIO_DD * pAccesDioDd);
STATUS      vxbResourceShow (int fd, VXB_RESOURCE * pVxbRes, char * name);

/* Functions =================================================================*/

/*******************************************************************************
*
* accesDioClose - close the DIO card and make it available to another user
*
* This function closes the card and makes it available to another task. It is
* not an error to close the card if it is not open. It is an error for a task
* to attempt to close an open the card if it is not the owner (that is, if it is
* not the task that opened it). Closing the card disables interrupts if enabled,
* and always clears any pending interrupts and resets the ISR callback function
* pointer to NULL.
*
* RETURNS: OK or ERROR per errno.
*
* ERRNO
*   ENXIO       Driver not installed.
*   EACCES      Card is open for another task.
*/

STATUS accesDioClose (void)
    {
    DBG_ENTER;

    if (!isDriverOK ()) { EXIT_ERR(ENXIO); return ERROR; }

    if (pAccesDioDd->taskId != taskIdSelf ())            /* Another owner? */
        { EXIT_ERR(EACCES); return ERROR; }

    /* With error checks passed, use the "force" version of this function
     * instead of repeating its code. */

    accesDioCloseForce ();

    DBG_EXIT_OK;
    return OK;
    }

/*******************************************************************************
*
* accesDioCloseForce - close the DIO card even if not the owner
*
* This function closes the card even if the calling task is not the owner. No
* error checks. It is intended for internal use and is not in the header file.
* See accesDioClose() for the user-callable "safe" version.
*
* RETURNS: N/A
*/

void accesDioCloseForce (void)
    {
    DBG_ENTER;

    /* First disable interrupts if enabled (which clears any pending interrupts),
     * and reset the ISR callback function pointer to NULL. Ignore any error from
     * accesDioIntDisable() since this is the "force" function. */

    (void) accesDioIntDisable ();
    errno = 0;                          /* In case accesDioIntDisble() failed. */
    pAccesDioDd->pIntCallback = NULL;

    /* "Close" the card by setting the taskId of the current "owner" (if any) to
     * NULL. (Do this last so other functions called in this function won't
     * fail.) */

    pAccesDioDd->taskId = 0;

    DBG_EXIT_OK;
    }

/*******************************************************************************
*
* accesDioConfigure - set the direction of each port to input or output
*
* Set the direction of each port either to get input from the pins or put output
* to the pins based on its corresponding argument. The arguments must be from
* the PORT_DIR enumeration defined in the vxbAccesDio.h header file
* (PORT_DIR_INPUT or PORT_DIR_OUTPUT). See the note regarding "port direction"
* and "read/write" to a port with the definition of PORT_DIR.
*
* NOTE: A port must be set to input direction to generate CoS interrupts. See
* accesDioIntEnable() for more.
*
* RETURNS: OK or ERROR per errno.
*
* ERRNO
*   ENXIO       Driver not installed.
*   ENODEV      Card is not open.
*   EACCES      Card is open for another task.
*/

STATUS accesDioConfigure
    (
    PORT_DIR    IOGroup0Dir,   /* I/O Group 0 direction */
    PORT_DIR    IOGroup1Dir    /* I/O Group 1 direction */
    )
    {
    UINT8       controlByte;    /* For debug but no harm if non-debug. */
    /*
     * The direction for each port is controlled by a single bit in the Control
     * Word byte. For each port, 0 means output and 1 means input. The bits in the
     * Control Word byte are:
     *
     *      ---------------------------------------------------
     *      | D7  | D6  | D5  | D4  | D3  | D2  | D1    | D0   |
     *      ---------------------------------------------------
     *      |  0  |  0  |  0  |  0  |  0  |  0  | 1dir  | 0dir |
     *      ---------------------------------------------------
     * where "ndir" is the control bit for I/O Group n; 1 == output 
     */
    DBG_MSG (1, "Enter: port0Dirs: 0=%d, port1Dirs=%d\n",
    		IOGroup0Dir, IOGroup1Dir);

    if (isEntryError (pAccesDioDd)) return ERROR;

    controlByte = (UINT8)( ((IOGroup0Dir == PORT_DIR_OUTPUT) ? 0x01 : 0x00));
    controlByte |= (UINT8)( ((IOGroup1Dir == PORT_DIR_OUTPUT) ? 0x02 : 0x00));

#ifdef ACCES_DEBUG
    DBG_MSG (2, "Write Control Byte @ I/O address 0x%x=0x%02hhx\n",
                   pAccesDioDd->bar + REG_TTLCONTROL, controlByte);
#endif /* ACCES_DEBUG */

    vxbWrite8 (pAccesDioDd->handle, (UINT8 *)pAccesDioDd->bar + REG_TTLCONTROL,
               controlByte);

    DBG_EXIT_OK;
    return OK;
    }

#ifdef ACCES_DEBUG
/*******************************************************************************
*
* accesDioDebugLevelSet - set verbosity for debug output
*
* Set accesDioDebugLevel to control debug output verbosity. See "Debugging"
* topic above for valid values.
*
* RETURNS: N/A
*/

void accesDioDebugLevelSet (int verbosityLevel)
    {
    accesDioDebugLevel = verbosityLevel;
    }

#endif /* ACCES_DEBUG */
/******************************************************************************
*
* accesDioGetDriverRevision - show driver revision and build date/time
*
* Show the current driver and device details for the single DIO card supported
* in this version of this driver.  Does not require hardware. Does not generate 
* errors.
*
* RETURNS: N/A
*
* ERRNO: N/A
*/
void accesDioGetDriverRevision (void)
    {
	printf ("mPCIe-IDIO-8 Driver Revision %s %s %s\n", ACCES_IDIO_DRIVER_REV, __DATE__, __TIME__);
    }


/*******************************************************************************
*
* accesDioInt - interrupt service routine to handle an interrupt
*
* Handle an interrupt from the ACCES DIO card. Call the user callback function
* if:
* - Interrupts are enabled (by a prior call to accesDioIntEnable()).
* - The current value of the ISR callback function pointer is not NULL. The
*   callback ISR pointer is set by calling accesDioIntCallbackInstall().
* - The card is open. (NOTE: the task ID is NOT checked--any task may be running
*   when an interrupt occurs, not just the task that opened the card).
*
* Whether a user callback function is called or not, always clear the interrupt
* at the end of this ISR (otherwise we would immediately receive another
* interrupt).
*
* RETURNS: If driver initialization, probe, or attach failed, return ERROR per
* errno. If the card is not open, interrupts are disabled, or there is no
* non-NULL callback function pointer, return OK. Else return the value returned
* by the callback function. (NOTE: Error ENXIO returned by the ISR cannot be
* distinguished from ENXIO returned by the driver check in this function.)
*
* ERRNO
*   ENXIO       Driver not installed.
*
* *****************************************************************************
* * CAUTION: If interrupts are enabled, generating debug output may cause the *
* * system to crash after a few interrupts. See this caution above under the  *
* * "Debugging" topic for how to avoid this.
* *****************************************************************************
*
* Note: The argument passed to this ISR when an interrupt occurs is set up by
* the call to vxbIntConnect() in our Attach function--it is the fourth
* argument passed to vxbIntConnect(). For other VxWorks drivers, it is usually
* either the VXB_DEV_ID passed to Probe and Attach, or the device descriptor
* (DD) for devices that create the DD dynamically. Since our DD accesDioDd is
* static, we call vbxIntConnect() with NULL for the fourth argument and this
* function can ignore it.
*/

LOCAL STATUS accesDioInt (void * ignore)
    {
    STATUS status;

    pAccesDioDd->intCount++;    /* For debug but generally useful. */

    DBG_ENTER;

    if (pAccesDioDd == NULL) { EXIT_ERR(ENXIO); return ERROR; }

    if ( pAccesDioDd->intEnabled &&
        (pAccesDioDd->pIntCallback != NULL) &&
        (pAccesDioDd->taskId != 0))
        {
        status = (pAccesDioDd->pIntCallback) ();

        DBG_MSG (1, "Exit. status=%i\n", status);
        }
    else
        {
        status = OK;
        DBG_MSG (1, "Exit. No user callback function installed.\n");
        }

    /* Clearing the interrupt should be the *Last* action of any ISR. */
    vxbWrite8 (pAccesDioDd->handle,
               (UINT8 *)pAccesDioDd->bar + REG_IRQ_CLR_STAT, 0xFF);

    return status;
    }

/*******************************************************************************
*
* accesDioIntCallbackInstall - install the user interrupt service routine (ISR)
*
* Record the callback interrupt service routine (ISR) function pointer argument.
* The prototype for that function must be:
*
*   STATUS <ISR function name> (void)
*
* The device must be open by calling accesDioOpen() before calling this routine.
*
* The ISR will be called if:
* - This routine has been called to establish the ISR callback function pointer.
* - Interrupts are enabled by calling accesDioIntEnable().
* - There is a change of state (CoS) on a pin on a port configured for input
*   while interrupts are enabled.
*
* CoS will be ignored if interrupts are not then enabled, and are ignored for
* pins on a port set for output direction.
*
* This routine may be called at any time that the card is open to change the
* callback function pointer. It may be called with NULL function pointer to
* remove the callback function.
*
* RETURNS: OK or ERROR per errno.
*
* ERRNO:
*   ENXIO       Driver not installed.
*   ENODEV      Card is not open.
*   EACCES      Card is open for another task.
*/

STATUS accesDioIntCallbackInstall
    (
    STATUS  (*callback_funcptr) (void)   /* callback function pointer */
    )
    {
    DBG_MSG (1, "Enter. callback_funcptr=%p\n", callback_funcptr);

    if (isEntryError (pAccesDioDd)) return ERROR;

    pAccesDioDd->pIntCallback = callback_funcptr;

    DBG_EXIT_OK;
    return OK;
    }

/*******************************************************************************
*
* accesDioIntClear - clear a pending interrupt, if any
*
* If an interrupt is pending, this function clears it; otherwise there is no
* effect. NOTE: It is unusual for a user task to call this function because
* interrupts are cleared as part of the interrupt service routine.
*
* RETURNS: OK or ERROR per errno.
*
* ERRNO
*   ENXIO       Driver not installed.
*   ENODEV      Card is not open.
*   EACCES      Card is open for another task.
*/

STATUS accesDioIntClear (void)
    {
    DBG_ENTER;

    if (isEntryError (pAccesDioDd)) return ERROR;

    vxbWrite8 (pAccesDioDd->handle,
               (UINT8 *)pAccesDioDd->bar + REG_IRQ_CLR_STAT, 0xFF);

    DBG_EXIT_OK;
    return OK;
    }

/*******************************************************************************
*
* accesDioIntDisable - disable interrupts from the card
*
* This routine disables interrupts from the card. Any pending interrupts are
* cleared. Calling this routine when interrupts are disabled is not an error.
*
* RETURNS: OK or ERROR per errno.
*
* ERRNO
*   ENXIO       Driver not installed.
*   ENODEV      Card is not open.
*   EACCES      Card is open for another task.
*
* Implementation Note: See accesDioIntEnable() enable/disable mechanism.
*/

STATUS accesDioIntDisable (void)
    {
    DBG_ENTER;

    if (isEntryError (pAccesDioDd)) return ERROR;

    /* Tests passed. Disable interrupts and update the device descriptor. Then
     * clear any pending interrupts. */

    vxbWrite8 (pAccesDioDd->handle,
               (UINT8 *)pAccesDioDd->bar + REG_IRQ_EN, 0);
    
    pAccesDioDd->intEnabled = FALSE;

    vxbWrite8 (pAccesDioDd->handle,
               (UINT8 *)pAccesDioDd->bar + REG_IRQ_CLR_STAT, 0xFF);

    DBG_EXIT_OK;
    return OK;
    }

/*******************************************************************************
*
* accesDioIntEnable - enable interrupts from the card
*
* This routine enables Change-of-State (CoS) interrupts from the card. A CoS
* interrupt occurs whenever any input pin changes state due to an external
* event. Calling this function when interrupts are enabled is not an error.
* Interrupts are not generated for ports set for output.
*
* RETURNS: OK or ERROR per errno.
*
* ERRNO
*   ENXIO       Driver not installed.
*   ENODEV      Card is not open.
*   EACCES      Card is open for another task.
*
* Implementation Note
* -------------------
* To get Change of State (CoS) interrupts requires five steps plus one:
* (1) Call vxbIntConnect(). Done in our Attach function.
* (2) Call vxbIntEnable(). Done in our Attach function. Could toggle this and
*     vxbIntDisable() to control interrupts, but do not--as far as VxWorks is
*     concerned, the interrupts are permanently enabled. Thus, vxbIsrVecsShow()
*     will always show [CONNECTED] and [ENABLED] for our DIO interrupt; use
*     accesDioShow() in this driver to show whether CoS interrupts are enabled.
* (3) At the card level, write 0xf8 to REG_COS_MASK. Could be done at the end 
*     of Attach (but not before calling iosDevAdd(), but it is best to leave 
*     the REG_COS_MASK register in its initial state if interrupts are not used.
* (4) The user must call accesDioConfigure() with argument PORT_DIR_INPUT for
*     each port on which interrupts are wanted. Note: This is the reset-default 
*     state of each port, thus is only necessary if they've been set to output.
* (5) At the card level, write 0xff to REG_IRQ_EN--the purpose of this function;
*     writing 0x00 disables the CoS interrupts. Use this mechanism here and in
*     accesDioIntDisable() to control whether interrupts are enabled or not.
*
* (6) The user must call accesDioIntCallbackInstall(); while this step is not
*     required, if the five steps above are done, the interrupt will occur, the
*     ISR will be invoked, and interrupt cleared--but there will be no effect
*     on user tasks beyond slight delays.
*/

STATUS accesDioIntEnable (void)
    {
    DBG_ENTER;

    if (isEntryError (pAccesDioDd)) return ERROR;
    
     vxbWrite8 (pAccesDioDd->handle,
                   (UINT8 *)pAccesDioDd->bar + REG_IRQ_EN, 0xff);

    pAccesDioDd->intEnabled = TRUE;

    DBG_EXIT_OK;
    return OK;
    }



/*******************************************************************************
*
* accesDioOpen - open the DIO card and assign it to the user
*
* This version of this driver supports a single DIO card in the system. This
* function opens that card and assigns it to the calling user task. A DIO card
* must be opened before it can be used. It stays open for the exclusive use of
* the calling user until it is freed by calling acessDioClose().
*
* It is an error to call this function if the card is already opened for either
* the current task or another task.
*
* RETURNS: OK or ERROR per errno.
*
* ERRNO
*   ENXIO       Driver not installed.
*   ENODEV      Card is already open for this task.
*   EACCES      Card is open for another task.
*/

STATUS accesDioOpen (void)
    {
    DBG_ENTER;

    if (pAccesDioDd == NULL) { EXIT_ERR(ENXIO); return ERROR; }

    if (pAccesDioDd->taskId == 0)
        {
        pAccesDioDd->taskId = taskIdSelf ();
        DBG_EXIT_OK;
        return OK;
        }

    /* Card already open. Set errno depending on whether this task has it or
     * another task does. */

    if (pAccesDioDd->taskId == taskIdSelf ())
        EXIT_ERR(ENODEV);
    else
        EXIT_ERR(EACCES);

    return ERROR;
    }

/*******************************************************************************
*
* accesDioPciAttach - attach the ACCES DIO card during PCI bus enumeration
*
* Attach an ACCES DIO card. Our DIO device matched its probe so this Attach
* function is called. Its job is to add the DIO *device* to the I/O sub-system.
* After this function completes successfully, the device is ready to be opened,
* read/written, etc. (by calls to our device-specific accesDioOpen(), etc.).
*
* For context in understanding this function, here is our device descriptor
* structure. See "DIO Device Descriptor and Pointer to It" above for details.
*/
#if 0
typedef struct accesDioDd
    {
    DEV_HDR         devHdr;             /* I/O device header; includes device
                                           name and driver number */
    char *          pProductName;       /* ACCES product name; set by
                                           accesDioAttach() to the name in the
                                           DIO_PCI_DEV_CFG structure
                                           (see "Match Table" below); example:
                                           "mPCIe-DIO-24S" */
    TASK_ID         taskId;             /* ID of task card was opened by; set by
                                           accesDioOpen(); NULL means closed.*/
    ULONG           bar;                /* BAR virtual address */
    void *          handle;             /* see "handle" above */
    STATUS          (*pIntCallback)();  /* for pointer to user function to call
                                           on an interrupt */
    BOOL            intEnabled;         /* TRUE if interrupts are enabled */
    VXB_RESOURCE *  pAdrRes;            /* address resource described above */
    VXB_RESOURCE *  pIntRes;            /* interrupt resource described above */

    } ACCES_DIO_DD;
#endif
/*
* To add the (single) DIO device to the I/O subsystem, this function:
*
* - Points our module LOCAL pAccesDioDd at our module LOCAL device descriptor
*   accesDioDd struct (it's initial values will be all zero--sufficient for this
*   case). (NOTE: A VxBus driver will often allocate the device descriptor
*   dynamically for each instance of a kind of device, perhaps using a "create"
*   function for the type of device (e.g., a serial terminal). Because this
*   version supports only one DIO card in the system and uses a static device
*   descriptor, no "create" function is needed, but because this pointer
*   is to the *device* descriptor, it is more correct to set it in this device
*   Attach function than in the accesDioPciDrvInit() function).
*
* - Uses an "address resource" to get the virtual address of our device (it's
*   "bar", which maps to the devices physical Base Address Register, BAR) and
*   a related "handle" (accessing a device register requires the handle, the
*   bar, and the offset of the register from the bar) and stores the bar, the
*   handle, and the pointer to the address resource in our device descriptor.
*   See pAdrRes under topic "DIO Device Descriptor and Pointer to It" above
*   for details.
*
* - Gets and initializes an "interrupt resource" for interrupts, then connects
*   and enables them. See pIntRes under topic "DIO Device Descriptor and Pointer
*   to It" above for details.
*
* - Saves a pointer to product name from the configuration data in our device
*   descriptor.
*
* - Finally, call iosDevAdd() to add our DIO device to the I/O subsystem.
*   (Calling it at the end avoids having to delete the device on any failure
*   of the above steps.)
*
* RETURNS: OK or ERROR per ERRNO.
*
* ERRNO
*   ENODEV      Driver is already installed--invalid to call this function.
*   ENOMEM      Resource allocation failed (BAR address or interrupt).
*   ENXIO       Driver init error (not done or Probe error) or failure in
*               this function other than the above (e.g., cannot set up
*               interrupts).
*   EIO         Problem with BAR or IRQ (e.g., BAR is invalid 0).
*/

LOCAL STATUS accesDioPciAttach
    (
    struct vxbDev * pVxbDev     /* VxBus device struct, NOT our device descriptor */
    )
    {
    DIO_PCI_DEV_CFG *   pAccesDioPciDevCfg;
    VXB_RESOURCE_ADR *  pAdrRes_pRes;
    VXB_RESOURCE_ADR *  pBAR0Res_pRes = NULL;
    DEV_HDR *           pDevHdr;            /* For convenience. See below. */

#ifdef ACCES_DEBUG
    accesDio_pVxbDev = pVxbDev;
#endif /* ACCES_DEBUG */

    kprintf("attach1 pVxbDev=%p\n", pVxbDev); /* J2H Debug */
    DBG_MSG (1, "Enter. pVxbDev=%p\n", pVxbDev);

    /* With only one card in the system, only allow a single attach.
    * (Will not be invalid if support multiple cards.) */
    if (drvNum > 0) { EXIT_ERR(ENODEV); kprintf("attach1.5err\n"); /* J2H Debug */ return ERROR; }

    /* Install the driver with all NULL arguments as explained above. */
    drvNum = iosDrvInstall (NULL, NULL, NULL, NULL, NULL, NULL, NULL);
    if (drvNum == ERROR)
        {kprintf("attach1.75 err\n"); /* J2H Debug */ 
        DBG_MSG (1, "iosDrvInstall() call to install driver failed.\n");
        return ERROR;
        }
    else
        DBG_MSG (2, "iosDrvInstall() OK\n");

    /* Point the pAccesDioDd pointer used throughout this driver to our module
     * LOCAL device descriptor at our module LOCAL device descriptor. For
     * convenience, point pDevHdr at its first member, the DEV_HDR.
     */

    pAccesDioDd = & accesDioDd;
    pDevHdr = (DEV_HDR *) pAccesDioDd;

    /* Get and initialize a resource for an I/O-type device (2nd argument)
     * for its BAR[2] (the 3rd argument is the BAR #). Then save the handle in
     * two steps from the resource, the bar from its virtual address, and
     * the pointer to the address resource in our device descriptor. */

    pAccesDioDd->pAdrRes = vxbResourceAlloc (pVxbDev, VXB_RES_IO, 2);
    if (pAccesDioDd->pAdrRes == NULL) 
    { 
    	kprintf("attach2 failed BAR2 alloc"); // J2H Debug
    	EXIT_ERR(ENOMEM); 
    	goto cleanup; 
    }

    DBG_MSG  (2, "alloc pAccesDioDd->pAdrRes = VXB_RES_IO = %p\n", pAccesDioDd->pAdrRes);

    pAdrRes_pRes = (VXB_RESOURCE_ADR *)pAccesDioDd->pAdrRes->pRes;
    if (pAdrRes_pRes == NULL) { EXIT_ERR(ENOMEM); goto cleanup; }

    pAccesDioDd->handle = pAdrRes_pRes->pHandle;
    pAccesDioDd->bar = (ULONG)pAdrRes_pRes->virtual;

    DBG_MSG (2, "got handle %p and bar 0x%lx for BAR[2]\n",
                   pAccesDioDd->handle, pAccesDioDd->bar);

    /*
     * A bar of 0 is invalid and suggests a hardware problem (e.g., card or
     * cable not well-seated). Detect this and fail.
     */
    if (pAccesDioDd->bar == 0) { EXIT_ERR(EIO); kprintf("attach2.5err\n"); /* J2H Debug */ goto cleanup; }


    /* Get and initialize a resource for an I/O-type device (2nd argument)
     * for its BAR[0] (the 3rd argument is the BAR #). Then save the handle in
     * two steps from the resource, the bar from its virtual address, and
     * the pointer to the address resource in our device descriptor. */

    pAccesDioDd->pBAR0Res = vxbResourceAlloc (pVxbDev, VXB_RES_IO, 0);
    if (pAccesDioDd->pBAR0Res == NULL) 
    {
    	kprintf("attach3 failed BAR1 alloc"); // J2H Debug
    	EXIT_ERR(ENOMEM); 
    	goto cleanup; 
    } 

    DBG_MSG (2, "alloc pAccesDioDd->pBAR0Res = VXB_RES_IO = %p\n", pAccesDioDd->pBAR0Res);

    if (pAccesDioDd->pBAR0Res != NULL)
    	pBAR0Res_pRes = (VXB_RESOURCE_ADR *)pAccesDioDd->pBAR0Res->pRes;
    if (pBAR0Res_pRes == NULL) { EXIT_ERR(ENOMEM); goto cleanup; } 

    if (pBAR0Res_pRes != NULL)
    {
    	pAccesDioDd->handle0 = pBAR0Res_pRes->pHandle;
    	pAccesDioDd->BAR0 = (ULONG)pBAR0Res_pRes->virtual;
    }

    DBG_MSG (2, "got handle %p and bar 0x%lx for BAR[0]\n",
                   pAccesDioDd->handle0, pAccesDioDd->BAR0);

    /*
     * A bar of 0 is invalid and suggests a hardware problem (e.g., card or
     * cable not well-seated). Detect this and fail.
     */
    if (pAccesDioDd->BAR0 == 0) { EXIT_ERR(EIO); goto cleanup; } 



    /*
     * Get and initialize a resource for the Change of State interrupts, and
     * connect and enable them at the VxWorks driver level and partly at the
     * card level (see accesDioIntEnable() for more).
     *
     * The third argument to vxbIntConnect() is our ISR, and fourth argument
     * is a pointer to an argument to be passed to our ISR, accesDioInt().
     * See it for the rationale for the fourth argument.
     *
     * Three more steps are required after this to get interrupts. See
     * accesDioIntEnable() for details.
     */

    pAccesDioDd->pIntRes = vxbResourceAlloc (pVxbDev, VXB_RES_IRQ, 0);
    if (pAccesDioDd->pIntRes == NULL) { EXIT_ERR(ENOMEM); goto cleanup; }

    DBG_MSG (2, "alloc pIntRes = VXB_RES_IRQ =%p\n", pAccesDioDd->pIntRes);

    if (vxbIntConnect (pVxbDev, pAccesDioDd->pIntRes, (VOIDFUNCPTR) accesDioInt, NULL)
            != OK)
        { EXIT_ERR(ENXIO); goto cleanup; }

    DBG_MSG (2, "vxbIntConnect() OK.\n");

    if (vxbIntEnable (pVxbDev, pAccesDioDd->pIntRes) != OK)
        { EXIT_ERR(ENXIO); goto cleanup; }

    DBG_MSG (2, "vxbIntEnable() OK.\n");
   
    /* Save pointer to product name from the configuration data in our device
     * descriptor. */

    pAccesDioPciDevCfg = (DIO_PCI_DEV_CFG *)vxbDevDrvDataGet (pVxbDev);
    if (pAccesDioPciDevCfg == NULL || pAccesDioPciDevCfg->pProductName == NULL)
        { EXIT_ERR(ENXIO); goto cleanup; }

    pAccesDioDd->pProductName = pAccesDioPciDevCfg->pProductName;

    /* Finally, call iosDevAdd() to add our DIO device to the I/O subsystem.
     * Arguments:
        * - Pointer to the DEV_HDR for the device (local pDevHdr, set above--
        *   the first member of our accesDioDd device descriptor).
        * - Hard-coded device name "/accesDio/0". When extended to multiple
        *   DIO cards, a local static variable in this file could be initialized
        *   in accesDioPciDrvInit() and incremented for each device as it is
        *   attached.
        * - The driver number returned by iosDrvInstall() called from our
        *   accesDioPciAttach() function and stored in module-level static
        *   drvNum.
     */
     if (iosDevAdd (pDevHdr, "/accesDio/0", drvNum) == ERROR)
         { EXIT_ERR(ENXIO); goto cleanup; }

    DBG_MSG (2, "iosDevAdd() OK.\n");
    
    sysOutByte(pAccesDioDd->BAR0 + 0x69, 9); /* enable PCIe bridge IRQ like PEX8311 */
    
    kprintf("attach9.99 success\n"); /* J2H Debug */ 
    
    DBG_EXIT_OK;
    return OK;

    /*
     * Common exit for errors that require freeing memory allocated here. errno
     * should be set and the "Exit ERROR" message already issued before jumping
     * here.
     */

cleanup:
    if (pAccesDioDd->pAdrRes != NULL)
        (void) vxbResourceFree (pVxbDev, pAccesDioDd->pAdrRes);

    if (pAccesDioDd->pIntRes != NULL)
        {
        (void)vxbResourceFree (pVxbDev, pAccesDioDd->pIntRes);
        }

    return ERROR;
    }

/*******************************************************************************
*
* accesDioPciDrvInit - initialize the ACCES DIO driver
*
* This function serves three purposes:
*
* 1. It provides a placeholder for partial initialization of this driver. This
*    function is called during VxWorks initialization by VIP:prjConfigs.c:
*    usrRoot()) when component DRV_ACCES_DIO_PCI is configured into VxWorks
*    because of the INIT_RTN and _INIT_ORDER statements in our component
*    description file 40vxbAccesDio.cfd.
*
*    A driver initialization function might call iosDrvInstall() to register the
*    driver with the I/O system, might set an initial unit number, might take
*    other actions to support multiple instances of a devices. However, this
*    driver supports only one DIO card in a system initially, and this function
*    serves mainly purposes 2 and 3 below. (Calling iosDrvInstall() is deferred
*    to our Attach function to avoid complications related to when this function
*    is called--see Attach for details).
*
* 2. Because it is called during system initialization, the linker includes the
*    file in the image. If no function in the file is called, then the file is
*    not included in the image.
*
* 3. Sets up for debug output if ACCES_DEBUG is defined.
*
* RETURNS: N/A
*
* ERRNO N/A
*/

void accesDioPciDrvInit (void)
    {
    DBG_ENTER;
    DBG_EXIT_OK;
    }

/*******************************************************************************
*
* accesDioPciProbe - probe for device presence
*
* Respond to a probe from higher-level VxBus setup code for a device discovered
* on a PCI bus. Look in our "match" table for an entry with matching vendor and
* device ID. The argument to this function is a pointer to a vxbDev structure
* (to avoid confusion--not our device descriptor accesDioDd struct). It is
* passed from this Probe function to our Attach by the VxBus device enumeration
* process, and we can use a member set by calling vxbDevDrvDataSet() to pass a
* pointer to our configuration data.
*
* RETURNS: OK if a matching device is found, else ERROR. (VxBus supports a "best
* match" feature--see vxbLib.c:vxbDevProbe(); it is not used here.)
*
* ERRNO: N/A
*/

LOCAL STATUS accesDioPciProbe
    (
    struct vxbDev * pVxbDev     /* VxBus device struct, NOT our device descriptor */
    )
    {
#ifdef ACCES_DEBUG
    /* Set up for DBG_... statements by initializing the kwrite function to use
     * kwriteVga. */
    _func_kwrite = kwriteVga;  // --J2H debug
#endif /* ACCES_DEBUG */

    VXB_PCI_DEV_MATCH_ENTRY *pMatch;
    DIO_PCI_DEV_CFG   *pCfg;
#ifdef ACCES_DEBUG    	


    PCI_HARDWARE * pPciDevInfo; /* So can get Vendor and Device IDs. */

    accesDio_pVxbDev = pVxbDev; /* Save argument in module LOCAL for debug. */

    /*
     * If module LOCAL accesDioProbeFail is TRUE and accesDioDebugLevel >=3,
     * fail the probe immediately so that the DIO device is NOT discovered
     * during normal VxBus PCI enumeration. Can then call Probe and Attach
     * after VxWorks is up. Further details under "Debugging" above.
     */
 
    pPciDevInfo = (PCI_HARDWARE *)vxbDevIvarsGet (pVxbDev);

    kprintf("%04x:%04x ", pPciDevInfo->pciVendId, pPciDevInfo->pciDevId ); //J2H Debug

    // if ((accesDioDebugLevel >= 3) && accesDioProbeFail) return ERROR;    
#endif /* ACCES_DEBUG */

    
        
    /* In this case, ERROR means "not a match"; don't set errno. */
    if (vxbPciDevMatch (pVxbDev, matchTable, &pMatch) != OK)
        {
#ifdef ACCES_DEBUG
        if (pPciDevInfo->pciVendId == 0x494f) kprintf("no match.\n");
#endif /* ACCES_DEBUG */
        return ERROR;
        }


    /* Match! Save a pointer to the Match Table entry in the vxbDev argument for
     * our Attach function. */

    pCfg = (DIO_PCI_DEV_CFG *)pMatch->data; 
    vxbDevDrvDataSet (pVxbDev,(void *)pCfg);

    // DBG_MSG0 (1, "match!\n");
    return OK;
    }

/*******************************************************************************
* accesDioReadControl - read the Control Word (byte) from the device
*
* This function reads the Control Word from the device and stores it in the
* UINT8 location passed to it. While available, it would ordinarily not be
* called by a user.
*
* RETURNS: OK or ERROR per errno.
*
* ERRNO
*   ENXIO       Driver not installed.
*   ENODEV      Card is not open.
*   EACCES      Card is open for another task.
*   EINVAL      Null argument pointer.
*/

STATUS accesDioReadControl
    (
     UINT8 * pControl    /* pointer to location to store the Control Word */
    )
    {
    DBG_ENTER;

    if (isEntryError (pAccesDioDd)) return ERROR;
    if (pControl == NULL) { EXIT_ERR(EINVAL); return ERROR; }

    *pControl = (vxbRead8 (pAccesDioDd->handle,
                         (UINT8 *)pAccesDioDd->bar + REG_TTLCONTROL));

    DBG_MSG (1, "Exit OK: pControl = %#006x\n", *pControl);
    return OK;
    }

/*******************************************************************************
*
* accesDioReadPorts - read all Digital Bits, Isolated and TTL inputs/readbacks
*
* This function is passed a pointer to a UINT8 where the isolated input reading
* is stored.  It is passed a pointer to another UINT8 where the 8-bits of TTL
* digital bits are stored; either as read from bits if they have been configured
* for input mode, or as read-back from the output bits' control register latch
* for bits configured in output mode.
* 
* It is valid to pass NULL for either/both parameters; the function will simply
* not attempt to read the input type associated with a NULL buffer pointer.
*
* RETURNS: OK or ERROR per errno.
*
* ERRNO
*   ENXIO       Driver not installed.
*   ENODEV      Card is not open.
*   EACCES      Card is open for another task.
*   EINVAL      Null argument pointer.
*/

STATUS accesDioReadPorts
    (
    UINT8 *     pISOIn,     /* pointer to location to store the isolated input
                               values read from the device */
    UINT8 *     pTTLIn      /* pointer to location to store TTL inputs or the 
                                value read-back from the outputs */
    )

    {
    if (isEntryError (pAccesDioDd)) return ERROR;
    if (pISOIn != NULL) 
    	{
        *pISOIn = (vxbRead8 (pAccesDioDd->handle, 
        		         (UINT8 *)pAccesDioDd->bar + REG_ISODIG));
    	}

    if (pTTLIn != NULL) 
    	{
        *pTTLIn = (vxbRead8 (pAccesDioDd->handle,
                         (UINT8 *)pAccesDioDd->bar + REG_TTLDATA));
    	}
    return OK;
    }

/*******************************************************************************
* accesDioRead32Raw - return all "8255" registers (including Control Word)
*
* Read and return the "8255" registers as a UINT32 with no masking.  For
* internal use; no error checking.  Not in the header file--IMPORT explicitly to
* use.
*
* RETURNS: The "8255" registers as a UINT32.
*
* ERRNO: N/A
*/

UINT32 accesDioRead32Raw (void)
    {
    return vxbRead32 (pAccesDioDd->handle, (UINT32 *)pAccesDioDd->bar);
    }


/******************************************************************************
*
* accesDioShow - show driver and device details for the DIO card
*
* Show the current driver and device details for the single DIO card supported
* in this version of this driver. If the driver is not initialized or the card
* is not attached, the function reports that but does not return ERROR.
*
* For output, uses printf().
*
* RETURNS: N/A
*
* ERRNO: N/A
*/

void accesDioShow (void)
    {
    DEV_HDR *   pDevHdr = (DEV_HDR *)pAccesDioDd;

    printf ("\nACCES DIO Driver and Device Status\n");
    printf (  "----------------------------------\n");
    accesDioGetDriverRevision();
        
    if ((pAccesDioDd == NULL) || (pAccesDioDd->pProductName == NULL))
        {
        printf ("DIO driver not installed or Probe() or Attach() failed.\n");
        printf ("(device descriptor (accesDioDd) @ %p)\n", &accesDioDd);
        return;
        }

    printf ("driver number=%i\n", pDevHdr->drvNum);
    printf ("device descriptor (accesDioDd) @ %p\n", pAccesDioDd);
    printf ("device name=\"%s\"\n", pDevHdr->name);
    printf ("product name=\"%s\"\n", pAccesDioDd->pProductName);
    printf ("owner task ID (0 means not open)=%p\n", pAccesDioDd->taskId);
    printf ("BAR[2]=0x%lx\n", pAccesDioDd->bar);
    printf ("handle=%p\n", pAccesDioDd->handle);
    printf ("BAR[0]=0x%lx\n", pAccesDioDd->BAR0);
    printf ("handle0=%p\n", pAccesDioDd->handle0);
    printf ("interrupt callback funcptr=%p\n", pAccesDioDd->pIntCallback);
    printf ("interrupts enabled=%s\n", pAccesDioDd->intEnabled?"TRUE":"FALSE");
    printf ("pAdrRes=%p\n", pAccesDioDd->pAdrRes);
    printf ("pBAR0Res=%p\n", pAccesDioDd->pBAR0Res);
    printf ("pIntRes=%p\n", pAccesDioDd->pIntRes);
    printf ("intCount=%u\n", pAccesDioDd->intCount);
    }

/*******************************************************************************
*
* accesIdioWriteTTL - write to 8 TTL bits; writes to inputs are queued 1 deep
* and will be asserted if/when the port is configured for output
*
* RETURNS: OK or ERROR per errno.
*
* ERRNO
*   ENXIO       Driver not installed.
*   ENODEV      Card is not open.
*   EACCES      Card is open for another task.
*/

STATUS accesIdioWriteTTL
    (
    UINT8  ports           /* value to be written */
    )
    {
    DBG_MSG (1, "Enter: ports=%#010x\n", ports);

    if (isEntryError (pAccesDioDd)) return ERROR;

    accesIdioWriteTTLRaw (ports);

    DBG_EXIT_OK;
    return OK;
    }

/*******************************************************************************
* accesDioWrite8Raw - write any UINT8 to any device register
*
* Write any UINT8 to any device register.  For internal use, no error checking.
* Not in the header file--IMPORT explicitly to use.
*
* RETURNS: OK or ERROR per errno.
*
* ERRNO:
*/
STATUS accesDioWrite8Raw (REG reg, UINT8 value)
    {
    vxbWrite8 (pAccesDioDd->handle, (UINT8 *)pAccesDioDd->bar + reg, value);
    return OK;
    }

/*******************************************************************************
* accesIdioWriteTTLRaw - write all TTL/CMOS bits
*
* Write 8-bits of TTL data For internal use, no error checking.
* Not in the header file--IMPORT explicitly to use. 
*
* RETURNS: N/A
*/

void accesIdioWriteTTLRaw (UINT8 ports)
    {
	vxbWrite8 (pAccesDioDd->handle, (UINT8 *)pAccesDioDd->bar + REG_TTLDATA, ports);
    }


/*******************************************************************************
*
* accesIdioWriteFET - write to 8 FET bits
*
* RETURNS: OK or ERROR per errno.
*
* ERRNO
*   ENXIO       Driver not installed.
*   ENODEV      Card is not open.
*   EACCES      Card is open for another task.
*/

STATUS accesIdioWriteFET
    (
    UINT8  fets           /* value to be written */
    )
    {
    DBG_MSG (1, "Enter: fets=%#010x\n", fets);

    if (isEntryError (pAccesDioDd)) return ERROR;

    RelOutPortB (0, REG_FET, fets);

    DBG_EXIT_OK;
    return OK;
    }

/*******************************************************************************
*
* isDriverOK - test whether the driver is operating correctly
*
* Returns TRUE if the driver appears to be installed and set up, else FALSE.
*
* NOTE: Isolated in this function to allow for more complex tests of the driver
* in later versions.
*
* RETURNS: TRUE if the driver appears to be installed and operating correctly,
* else FALSE.
*/

LOCAL BOOL isDriverOK (void)
    {
    return ((pAccesDioDd != NULL) && (drvNum > 0) &&
            (pAccesDioDd->pProductName != NULL));
    }

/*******************************************************************************
*
* isEntryError - return TRUE if if any of several error checks fail
*
* Return TRUE if any of several checks on our LOCAL DIO device descriptor fail.
* Intended for use on entry to most user-callable functions in this file.
*
* Side effect: Sets errno based on the failing check per the below.
*
* RETURNS: TRUE if an error check fails; FALSE if there is no error. Thus, the
* typical first statement of a user-callable function in this file is:
*
*   if (isEntryError (pAccesDioDd)) return ERROR;
*
* ERRNO
*   ENXIO       Driver not installed.
*   ENODEV      Card is not open.
*   EACCES      Card is open for another task.
*/

LOCAL BOOL isEntryError (ACCES_IDIO_DD * pAccesDioDd)
    {
    /* Test for driver likely init, probe, or attach error. */
    if (!isDriverOK())
        { EXIT_ERR(ENXIO);  return TRUE; }

    /* Test for card not open. */
    if (pAccesDioDd->taskId == 0)
        { EXIT_ERR(ENODEV); return TRUE; }

    /* Test for card open for another task. */
    if (pAccesDioDd->taskId != taskIdSelf ())
        { EXIT_ERR(EACCES); return TRUE; }

    return FALSE;
    }

UINT32 RelInPortB(UINT32 DeviceNumber, UINT8 reg)
	{
	if (isEntryError (pAccesDioDd)) return AIO_Rel_Error_Sentinel;
	if (DeviceNumber) return AIO_Rel_Error_Sentinel;
	return (vxbRead8 (pAccesDioDd->handle, (UINT8 *)pAccesDioDd->bar + reg));
	}

UINT32 RelInPortW(UINT32 DeviceNumber, UINT8 reg)
	{
	if (isEntryError (pAccesDioDd)) return AIO_Rel_Error_Sentinel;
	if (DeviceNumber) return AIO_Rel_Error_Sentinel;
	return (vxbRead16 (pAccesDioDd->handle, (UINT16 *)pAccesDioDd->bar + reg));
	}

UINT32 RelInPortL(UINT32 DeviceNumber, UINT8 reg)
	{
	if (isEntryError (pAccesDioDd)) return AIO_Rel_Error_Sentinel;
	if (DeviceNumber) return AIO_Rel_Error_Sentinel;
	return (vxbRead32 (pAccesDioDd->handle, (UINT32 *)pAccesDioDd->bar + reg));
	}

UINT32 RelOutPortB(UINT32 DeviceNumber, UINT8 reg, UINT8 value)
	{
	if (isEntryError (pAccesDioDd)) return AIO_Rel_Error_Sentinel;
	if (DeviceNumber) return AIO_Rel_Error_Sentinel;
	
    vxbWrite8 (pAccesDioDd->handle, (UINT8 *)pAccesDioDd->bar + reg, value);
	return value;
	}

UINT32 RelOutPortW(UINT32 DeviceNumber, UINT8 reg, UINT16 value)
	{
	if (isEntryError (pAccesDioDd)) return AIO_Rel_Error_Sentinel;
	if (DeviceNumber) return AIO_Rel_Error_Sentinel;
	vxbWrite16 (pAccesDioDd->handle, (UINT16 *)pAccesDioDd->bar + reg, value);
/*	
    vxbWrite8 (pAccesDioDd->handle, (UINT8 *)pAccesDioDd->bar + reg, value & 0x00ff);
    vxbWrite8 (pAccesDioDd->handle, (UINT8 *)pAccesDioDd->bar + reg + 1, value >> 8);
*/    
	return value;
	}

UINT32 RelOutPortL(UINT32 DeviceNumber, UINT8 reg, UINT32 value)
	{
	if (isEntryError (pAccesDioDd)) return AIO_Rel_Error_Sentinel;
	if (DeviceNumber) return AIO_Rel_Error_Sentinel;
	vxbWrite32 (pAccesDioDd->handle, (UINT32 *)pAccesDioDd->bar + reg, value);	
/*    
	vxbWrite8 (pAccesDioDd->handle, (UINT8 *)pAccesDioDd->bar + reg, value & 0x000000FF);
    vxbWrite8 (pAccesDioDd->handle, (UINT8 *)pAccesDioDd->bar + reg + 1, (value >> 8) & 0x000000FF);
    vxbWrite8 (pAccesDioDd->handle, (UINT8 *)pAccesDioDd->bar + reg + 2, (value >> 16) & 0x000000FF);
    vxbWrite8 (pAccesDioDd->handle, (UINT8 *)pAccesDioDd->bar + reg + 3, (value >> 24) & 0x000000FF);
*/    
	return value;
	}

			
