/* vxbAccesDio.h - ACCES I/O Products DIO driver header file*/

/* Copyright (c) 2018, ACCES I/O Products, Inc.
 * 10623 Roselle St, San Diego CA, 92121-1506 USA
 * +1-858-550-9559 | +1-800-326-1649 | sales@accesio.com
 *
 * The right to copy, distribute, modify, or otherwise make use of this software
 * may be pursuant to the terms of an applicable ACCES I/O Products, Inc.
 * (ACCES) license agreement, or in the absence of such agreement, is limited to
 * use with ACCES hardware products.
 */

/*
modification history
--------------------
02apr19,j2h  forked for mPCIe-IDIO-8 beta 1
20dec18,j2h  Fixed comment typo
21jun18,acc  First release.
*/

/*
DESCRIPTION
===========
This header file defines and documents the user interface for developers using
various ACCES I/O Product, Inc., digital I/O cards, on a PCI bus, including the
mPCIe-DIO-24S. It also introduces the features of the hardware and provides
references for the hardware and PCI. (NOTE: in this context, "user" refers to
a VxWorks task calling an API.)

This version of the driver supports:

- A single ACCES IDIO card in the system, with device name /accesDio/0. To access
  the device, a user task must call accesDioOpen() to reserve it, giving that
  task exclusive access until a call to accesDioClose(). Thus, no special
  considerations are required to avoid conflict among multiple tasks or CPUs.

- Kernel mode only; no RTP support.

References
----------
Further details for DIO devices may be found in four sources:

- This file vxbAIO_IDIO.h: #defines for port masks, port access macros, and
  comments for each user-callable functions prototype.

For information on PCI, the following Wikipedia article and related articles
are a useful introduction:

 - Wikipedia. "PCI configuration space." Last edited 6 March 2018.
   https://en.wikipedia.org/wiki/PCI_configuration_space.
*/

#ifndef __INCaccesIdioDrvh
#define __INCaccesIdioDrvh

#ifdef __cplusplus
extern "C" {
#endif

/* Includes */

#include <vxWorks.h>

/*******************************************************************************
* Types and Defines
* =================
*/

/*
 * Port direction values passed to accesDioConfigure(): a port must be set to
 * output direction before writing to it, and set to input direction before
 * reading the state of the pins.
 *
 * NOTE: The concepts of "port direction" and of "read/write" to a port using
 * the functions provided here must be considered together:
 * - If a port is set to input direction, a write to it is ignored.
 * - If a port is set to output direction, a read from it will return the value
 *   last written to it.
 */

typedef enum port_dir
    {
    PORT_DIR_INPUT,     /* set port to get data from external devices */
    PORT_DIR_OUTPUT     /* set port to send data to external devices */
    } PORT_DIR;

#define AIO_Rel_Error_Sentinel (0xAA55AA55)
#define ACCES_IDIO_DRIVER_REV "0.10"
    
/*******************************************************************************
* User-Callable Function Prototypes
* =================================
* This driver provides the following user-callable functions (presented
* alphabetically in standard VxWorks format). There are no higher-level
* functions such as open() and ioctl().
*/

/*******************************************************************************
*
* accesDioClose - close the DIO card and make it available to another user
*
* This function closes the card and makes it available to another task. It is
* not an error to close the card if it is not open. It is an error for a task
* to attempt to close an open the card if it is not the owner (that is, if it is
* not the task that opened it). Closing the card disables interrupts if enabled,
* and always clears any pending interrupts and resets the ISR callback function
* pointer to NULL.
*
* RETURNS: OK or ERROR per errno.
*
* ERRNO
*   ENXIO       Driver not installed.
*   EACCES      Card is open for another task.
*/

STATUS accesDioClose (void);

/*******************************************************************************
*
* accesDioConfigure - set the direction of each port to input or output
*
* Set the direction of each port either to get input from the pins or put output
* to the pins based on its corresponding argument. The arguments must be from
* the PORT_DIR enumeration defined in the vxbAccesDio.h header file
* (PORT_DIR_INPUT or PORT_DIR_OUTPUT). See the note regarding "port direction"
* and "read/write" to a port with the definition of PORT_DIR.

*
* RETURNS: OK or ERROR per errno.
*
* ERRNO
*   ENXIO       Driver not installed.
*   ENODEV      Card is not open.
*   EACCES      Card is open for another task.
*/

STATUS accesDioConfigure
    (
    PORT_DIR    IOGroup0Dir,   /* I/O Group 0 direction */
    PORT_DIR    IOGroup1Dir    /* I/O Group 1 direction */
    );

/******************************************************************************
*
* accesDioGetDriverRevision - show driver revision and build date/time
*
* Show the current driver and device details for the single DIO card supported
* in this version of this driver.  Does not require hardware. Does not generate 
* errors.
*
* RETURNS: N/A
*
* ERRNO: N/A
*/
void accesDioGetDriverRevision (void);

/*******************************************************************************
*
* accesDioIntCallbackInstall - install the user interrupt service routine (ISR)
*
* Record the callback interrupt service routine (ISR) function pointer argument.
* The prototype for that function must be:
*
*   STATUS <ISR function name> (void)
*
* The device must be open by calling accesDioOpen() before calling this routine.
*
* The ISR will be called if:
* - This routine has been called to establish the ISR callback function pointer.
* - Interrupts are enabled by calling accesDioIntEnable().
* - There is a change of state (CoS) on a pin on a port configured for input
*   while interrupts are enabled.
*
* CoS will be ignored if interrupts are not then enabled, and are ignored for
* pins on a port set for output direction.
*
* This routine may be called at any time that the card is open to change the
* callback function pointer. It may be called with NULL function pointer to
* remove the callback function.
*
* RETURNS: OK or ERROR per errno.
*
* ERRNO:
*   ENXIO       Driver not installed.
*   ENODEV      Card is not open.
*   EACCES      Card is open for another task.
*/

STATUS accesDioIntCallbackInstall
    (
    STATUS  (*callback_funcptr)()   /* callback function pointer */
    );

/******************************************************************************
*
* accesDioIntClear - clear a pending interrupt, if any
*
* If an interrupt is pending, this function clears it; otherwise there is no
* effect. NOTE: It is unusual for a user task to call this function because
* interrupts are cleared as part of the interrupt service routine.
*
* RETURNS: OK or ERROR per errno.
*
* ERRNO
*   ENXIO       Driver not installed.
*   ENODEV      Card is not open.
*   EACCES      Card is open for another task.
*/

STATUS accesDioIntClear (void);

/******************************************************************************
*
** accesDioIntDisable - disable interrupts from the card
*
* This routine disables interrupts from the card. Any pending interrupts are
* cleared. Calling this routine when interrupts are disabled is not an error.
*
* RETURNS: OK or ERROR per errno.
*
* ERRNO
*   ENXIO       Driver not installed.
*   ENODEV      Card is not open.
*   EACCES      Card is open for another task.
*/

STATUS accesDioIntDisable (void);

/******************************************************************************
*
* accesDioIntEnable - enable interrupts from the card
*
* This routine enables Change-of-State (CoS) interrupts from the card. A CoS
* interrupt occurs whenever any input pin changes state due to an external
* event. Calling this function when interrupts are enabled is not an error.
* Interrupts are not generated for ports set for output.
*
* RETURNS: OK or ERROR per errno.
*
* ERRNO
*   ENXIO       Driver not installed.
*   ENODEV      Card is not open.
*   EACCES      Card is open for another task.
*/

STATUS accesDioIntEnable (void);

/*******************************************************************************
*
* accesDioIntPending - test whether an interrupt is pending from the card
*
* This routine is passed a pointer to a byte. It sets the byte to 1 if an
* interrupt was generated and not yet cleared (pending) and 0 if not. Reading
* the interrupt status does not clear it; call accesDioIntClear() to clear the
* interrupt.
*
* RETURNS: OK or ERROR per errno.
*
* ERRNO
*   ENXIO       Driver not installed.
*   ENODEV      Card is not open.
*   EACCES      Card is open for another task.
*   EINVAL      Null argument pointer.
*
* NOTE: Because interrupts are cleared as part of the interrupt service routine
* in this driver, accesDioIntPending() could set *pIntStatus to TRUE only if one
* or more of the conditions required for the ISR to be invoked on an interrupt
* were not met.
*/

STATUS accesDioIntPending
    (
    BOOL *  pIntStatus   /* pointer to byte to store the interrupt status:
                            1 if an interrupt is pending, 0 if not */
    );

/*******************************************************************************
*
* accesDioOpen - open the DIO card and assign it to the user
*
* This version of this driver supports a single DIO card in the system. This
* function opens that card and assigns it to the calling user task. A DIO card
* must be opened before it can be used. It stays open for the exclusive use of
* the calling user until it is freed by calling acessDioClose().
*
* It is an error to call this function if the card is already opened for either
* the current task or another task.
*
* RETURNS: OK or ERROR per errno.
*
* ERRNO
*   ENXIO       Driver not installed.
*   ENODEV      Card is already open for this task.
*   EACCES      Card is open for another task.
*/

STATUS accesDioOpen (void);

/*******************************************************************************
* accesDioReadControl - read the Control Word (byte) from the device
*
* This function reads the Control Word from the device and stores it in the
* UINT8 location passed to it. While available, it would ordinarily not be
* called by a user.
*
* RETURNS: OK or ERROR per errno.
*
* ERRNO
*   ENXIO       Driver not installed.
*   ENODEV      Card is not open.
*   EACCES      Card is open for another task.
*   EINVAL      Null argument pointer.
*/

STATUS accesDioReadControl
    (
    UINT8 * pControl    /* pointer to location to store the Control Word */
    );

/*******************************************************************************
*
* accesDioReadPorts - read all Digital Bits, Isolated and TTL inputs/readbacks
*
* This function is passed a pointer to a UINT8 where the isolated input reading
* is stored.  It is passed a pointer to another UINT8 where the 8-bits of TTL
* digital bits are stored; either as read from bits if they have been configured
* for input mode, or as read-back from the output bits' control register latch
* for bits configured in output mode.
* 
* It is valid to pass NULL for either/both parameters; the function will simply
* not attempt to read the input type associated with a NULL buffer pointer.
*
* RETURNS: OK or ERROR per errno.
*
* ERRNO
*   ENXIO       Driver not installed.
*   ENODEV      Card is not open.
*   EACCES      Card is open for another task.
*   EINVAL      Null argument pointer.
*/

STATUS accesDioReadPorts
    (
    UINT8 *     pISOIn,     /* pointer to location to store the isolated input
                               values read from the device */
    UINT8 *     pTTLIn      /* pointer to location to store TTL inputs or the 
                                value read-back from the outputs */
    );


/******************************************************************************
*
* accesDioShow - show driver and device details for the DIO card
*
* Show the current driver and device details for the single DIO card supported
* in this version of this driver. If the driver is not initialized or the card
* is not attached, the function reports that but does not return ERROR.
*
* For output, uses _func_kwrite (likely either kwriteVga() or kwriteSerial())
* if it is setup, or printf() if not.
*
* RETURNS: N/A
*
* ERRNO: N/A
*/

void accesDioShow (void);

/*******************************************************************************
*
* accesDioWritePorts - write to all of Ports A, B, CHi, and CLo
*
* This function is passed a single UINT32; it writes the low-order 24 bits to
* Ports A, B, CHi, and CLo. Port A is 0x000000FF, Port B is 0x0000FF00, Port CLo
* is 0x000F0000, and Port CHi is 0x00F00000. The MSB of the argument is ignored.
* Masks PORT_A_MASK .. PORT_C_MASK (CHi and CLo together) are defined in header
* file vxbAccesDio.h.
*
* RETURNS: OK or ERROR per errno.
*
* ERRNO
*   ENXIO       Driver not installed.
*   ENODEV      Card is not open.
*   EACCES      Card is open for another task.
*/

STATUS accesIdioWriteTTL
    (
    UINT8  ports           /* value to be written */
    );

STATUS accesIdioWriteFET
    (
    UINT8  fets           /* value to be written */
    );

UINT32 RelInPortB(UINT32 DeviceNumber, UINT8 reg);
UINT32 RelInPortW(UINT32 DeviceNumber, UINT8 reg);
UINT32 RelInPortL(UINT32 DeviceNumber, UINT8 reg);
UINT32 RelOutPortB(UINT32 DeviceNumber, UINT8 reg, UINT8 value);
UINT32 RelOutPortW(UINT32 DeviceNumber, UINT8 reg, UINT16 value);
UINT32 RelOutPortL(UINT32 DeviceNumber, UINT8 reg, UINT32 value);

/* the following two lines are convenience #defines to math the Windows API names */
#define RelInPort  RelInPortW
#define RelOutPort RelOutPortW

#ifdef __cplusplus
}
#endif

#endif  /* __INCaccesIdioDrvh */
